/*************************************************************************
 *                                                                       *
 *   B H C M i e   V 1 . 0                                  Sep. 2006    *
 *                                                                       *
 *   Reference: Bohren, Huffman,                                         *
 *              Absorption and Scattering of Light by Small Particles,   *
 *              Wiley-Intersciencen, 1983                                *
 *                                                                       *
 *   Author:    Sven Biermann                                            *
 *   Contact:   www.SvenBiermann.de                                      *
 *                                                                       *
 *************************************************************************/

#include <iostream>
#include <iomanip>
#include "numerical.h"
using namespace std;

void CalcMie(const double x, const cmplx m, const double *theta, const long nang, cmplx* S1, cmplx* S2);
void showHelp();
int main(int argc, char *argv[], char *envp[])
{
	const double pi_div_180 = 3.1415926535897932385/180.0;
	if (argc<6)  showHelp();
	int argvOffset = 1;
	double x = 0.0;  // Mie parameter
	cmplx refrel(0.0);  // relative refractive index
	if (strcmp(argv[1], "-d") == 0)
	{
		if (argc<9)  showHelp();
		double diameter   = atof(argv[2]);    // diameter of the sphere im m
		if (diameter <= 0.0)  showHelp();
		double wavelength = atof(argv[3]);    // wavelength in m
		if (wavelength <= 0.0)  showHelp();
		double refmed     = atof(argv[4]);    // refmed = (real) refractive index of surrounding medium
		if (refmed <= 0.0)  showHelp();
		double realrefsph = atof(argv[5]);    // real part of refractive index of the sphere
		if (realrefsph <= 0.0)  showHelp();
		cmplx  refsph(realrefsph, 0.0);       // refractive index of the sphere = real + i imaginary
		refrel = refsph/refmed;        // relative refractive index
		x = 3.1415926535897932385*diameter*refmed / wavelength;  // Mie parameter
		argvOffset = 6;
	}
	else
	{
  	x = atof(argv[1]);
		if (x <= 0.0)  showHelp();
		refrel.SetRe(atof(argv[2]));
		if (refrel.GetRe() <= 0.0)  showHelp();
		argvOffset = 3;
	}
	long   nang   = atol(argv[argvOffset]);   // number of angles between "FromDeg" and "ToDeg"
  if (nang <= 0)  showHelp();
	double theta1 = atof(argv[argvOffset+1]); // FromDeg
	if (theta1 < 0.0  ||  theta1 > 180.0)  showHelp();
	double theta2 = atof(argv[argvOffset+2]); // ToDeg
	if (theta2 < 0.0  ||  theta2 > 180.0)  showHelp();
	double dTheta = (theta2-theta1)*pi_div_180;  // convert to radian
  theta1 *= pi_div_180;
  argvOffset += 3;
	int dataType = 0;  // calculate intensities as standard option
	if (argvOffset  < argc-1)  showHelp();  // too many command line options
	if (argvOffset == argc-1)
	{
    if (strcmp(argv[argvOffset],"-a") == 0)
			dataType = 1;  // calculate scattering amplitudes
		else if (strcmp(argv[argvOffset],"-b") == 0)
			dataType = 2;  // calculate scattering amplitudes and intensities
		else if (strcmp(argv[argvOffset],"-c") == 0)
			dataType = 3;  // calculate scattering amplitudes, intensities, and phases
		else showHelp();
	}

	double *theta = new double[nang+1];  // Angles for CalcMieForGivenAngles(...)
	for (long j=1; j<=nang; j++)  theta[j] = theta1 + double(j-1)/double(nang-1)*dTheta;

	cmplx *S1, *S2;                // Scattering Amplitudes
	S1 = new cmplx[nang+1];        // Allocating Memory for S1 and S2  (CalcMieForGivenAngles(...))
	S2 = new cmplx[nang+1];

	cout << "Mie Scattering for  x = " << setprecision(32) << x << ",  m = " << refrel << endl;
	CalcMie(x, refrel, theta, nang, S1, S2);

	switch(dataType)
	{
	case 1:
	  cout << "Angle\tRe(S1)\tIm(S1)\tRe(S2)\tIm(S2)" << endl;
	  for (long j=1; j<=nang; j++)   // j<2*nang for CalcMie(...)
		  cout << theta[j] / pi_div_180 << "\t" << S1[j].GetRe() << "\t" << S1[j].GetIm() << "\t"
			                                      << S2[j].GetRe() << "\t" << S2[j].GetIm() << endl;
		break;
	case 2:
	  cout << "Angle\tRe(S1)\tIm(S1)\t|S1|^2\tRe(S2)\tIm(S2)\t|S2|^2" << endl;
	  for (long j=1; j<=nang; j++)   // j<2*nang for CalcMie(...)
		  cout << theta[j] / pi_div_180 << "\t"
			     << S1[j].GetRe() << "\t" << S1[j].GetIm() << "\t" << S1[j].AbsSqr() << "\t"
			     << S2[j].GetRe() << "\t" << S2[j].GetIm() << "\t" << S2[j].AbsSqr() << endl;
		break;
	case 3:
	  cout << "Angle\tRe(S1)\tIm(S1)\t|S1|^2\targ(S1)\tRe(S2)\tIm(S2)\t|S2|^2\targ(S2)" << endl;
	  for (long j=1; j<=nang; j++)   // j<2*nang for CalcMie(...)
		  cout << theta[j] / pi_div_180 << "\t"
			     << S1[j].GetRe() << "\t" << S1[j].GetIm() << "\t"
					 << S1[j].AbsSqr() << "\t" << S1[j].GetPhi() << "\t"
			     << S2[j].GetRe() << "\t" << S2[j].GetIm() << "\t"
					 << S2[j].AbsSqr() << "\t" << S2[j].GetPhi() << endl;
		break;
	default:
	  cout << "Angle\t|S1|^2\t|S2|^2" << endl;
	  for (long j=1; j<=nang; j++)   // j<2*nang for CalcMie(...)
		  cout << theta[j] / pi_div_180 << "\t" << S1[j].AbsSqr() << "\t" << S2[j].AbsSqr() << endl;
	}

	return 0;
}


void CalcMie(const double x, const cmplx m, const double *theta, const long nang, cmplx* S1, cmplx* S2)
{
	double *mu, *pi, *tau, *pi0, *pi1, psi0, psi1, psi;
	double apsi, apsi0, apsi1, chi, chi0, chi1, fn;
	cmplx  *D, y, xi, xi0, xi1, an, bn;
	long nstop, nmx, j, n;

	mu = new double[nang+1];
	pi = new double[nang+1];
	tau = new double[nang+1];
	pi0 = new double[nang+1];
	pi1 = new double[nang+1];

	double dang = 1.5707963267948966192 / double(nang - 1);

	y = m*x;
	nstop = long(x + 4.0*pow(x, 1.0/3.0) + 3.0);
	nmx = long(y.GetAbs());
	if (nstop>nmx)  nmx = nstop;
	nmx += 16;
	dang = 1.5707963267948966192 / double(nang-1);

	// Calculate mu = cos(theta) for every angle
	for (j=1; j<=nang; j++)  mu[j] = cos(theta[j]);

	// Calculate logarythmic derivative D[n] by downward recurrence starting with D[nmx]=0
	// where  D[n-1] = (n/y) - (1 / (D[n] + n/y))
	D = new cmplx[nmx+1];
	D[nmx] = cmplx(0.0,0.0);
	for (n=nmx; n>0; n--)
	{
		D[n-1] = (n/y) - (1.0 / (D[n] + n/y));
	}

	// Initializations
	for (j=1; j<=nang; j++)
	{
		pi0[j] = 0.0;
		pi1[j] = 1.0;
		S1[j] = cmplx(0.0, 0.0);
		S2[j] = cmplx(0.0, 0.0);
	}

	apsi0 = psi0 = cos(x);  chi1 = psi0;
	apsi1 = psi1 = sin(x);  chi0 =-psi1;
	xi0 = cmplx(psi0, -chi0);
	xi1 = cmplx(psi1, -chi1);
	n = 1;

	while (n <= nstop)
	{
		fn = (2.0*n+1.0) / (n*(n+1.0));  // Factor in S1- and S2-sums containing n only
		psi = (2.0*n-1.0)*psi1/x - psi0;
		apsi = psi;
		chi = (2.0*n-1.0)*chi1/x - chi0;
		xi = cmplx(apsi, -chi);

		an =   ((D[n]/m + n/x)*apsi - apsi1)
		     / ((D[n]/m + n/x)*xi - xi1);

		bn =   ((m*D[n] + n/x)*apsi - apsi1)
		     / ((m*D[n] + n/x)*xi - xi1);

		for (j=1; j<=nang; j++)
		{
			pi[j] = pi1[j];
			tau[j] = n*mu[j]*pi[j] - (n+1.0)*pi0[j];
			S1[j] += fn*(an*pi[j] + bn*tau[j]);
			S2[j] += fn*(an*tau[j] + bn*pi[j]);
		}

		psi0 = psi1;
		psi1 = psi;
		apsi1 = psi1;
		chi0 = chi1;
		chi1 = chi;
		xi1 = cmplx(apsi1, -chi1);
		n++;

		for (j=1; j<=nang; j++)
		{
			pi1[j] = ((2.0*n-1.0)/(n-1.0))*mu[j]*pi[j];
			pi1[j] = pi1[j] - n*pi0[j]/(n-1.0);
			pi0[j] = pi[j];
		}
	}
}


void showHelp()
{
	cout << "Too few, too many, or wrong command line options!" << endl;
	cout << "BHCMie calculates Mie scattering intensities or amplitudes." << endl;
  cout << "Usage:" << endl;
	cout << "  BHCMie x m nAngles FromDeg ToDeg       or" << endl;
	cout << "  BHCMie -d diameter wavelength nmed nsphere nAngles FromDeg ToDeg" << endl;
	cout << "where x is the Mie parameter, m is the (real) relative refractive index," << endl;
	cout << "nAngleg is the number of angles to be calculated between FromDeg and" << endl;
	cout << "ToDeg. You have to specify real values for each parameter except for" << endl;
	cout << "the number of angles which has to be an integer." << endl;
	cout << "The option -d as first parameter indicates that you specify the" << endl;
	cout << "diameter and (vacuum) wavelength in microns, the refractive index of" << endl;
	cout << "the medium and that of the sphere, respectively, instead of x and m." << endl;
	cout << "These parameters have to be real numbers." << endl;
	cout << "Further options may follow the above parameter list:"<< endl;
	cout << "  -a    calculates complex scattering amplitudes instead of intensities" << endl;
	cout << "  -b    calculates both" << endl;
	cout << "  -c    calculates both and additionally the phases (in rad)" << endl;
	cout << "BHCMie [parameters] >> Filename.txt    saves the output to a file" << endl;
	cout << "named \"FileName.txt\". If you want to use spaces within that name" << endl;
	cout << "you have to enclose it with quotation marks." << endl;
	cout << "Examples:" << endl;
	cout << "  BHCMie 1203.5 0.75 1280 0.0 30.5" << endl;
	cout << "  BHCMie -d 950.0 0.5145 1.333 1.0 1024 33.5 35.5 -b >> scattering.txt" << endl;
	exit(0);
}
